﻿using Grasshopper.Kernel;
using Rhino.Runtime.InProcess;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.Reflection;

namespace Tester
{
    class Program
    {
        // Note: Doesn't seem to be able to get RhinoCore in DEBUG mode in this example (throws exception), only Release.

        static Program()
        {
            //System.IO.Directory.SetCurrentDirectory(Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ProgramFiles), "Rhino 7", "System")); // DEBUG
            System.IO.Directory.SetCurrentDirectory(Path.Combine(Environment.ExpandEnvironmentVariables("%ProgramW6432%"), "Rhino 7", "System")); // RELEASE

            AddAssemblyResolveEvents();
        }
        
        static void Main(string[] args)
        {
            Solve(@"C:\Users\seli\Documents\Rhino\Sweep2Crash.gh"); // CHANGE ME!
        }
        
        public static void Solve(string definitionPath)
        {
            if (!LaunchRhinoGH())
            {
                return;
            }

            var arch = new GH_IO.Serialization.GH_Archive();

            arch.ReadFromFile(definitionPath);
            
            using (var doc = new GH_Document())
            {
                arch.ExtractObject(doc, "Definition");
                
                var allComps = doc.Objects.Where(o => o != null && o is GH_Component && !(o as GH_Component).Locked).Cast<GH_Component>();

                foreach (var comp in allComps)
                {
                    comp.CollectData();
                    comp.ComputeData(); // Here the sweep2 throws an exception
                }
            }
        }


        private static RhinoCore _rhinoCore;
        private static Grasshopper.Plugin.GH_RhinoScriptInterface _grasshopperInstance = null;
        public static bool LaunchRhinoGH()
        {
            try
            {
                _rhinoCore = new RhinoCore(new string[] { "/NOSPLASH" }, Rhino.Runtime.InProcess.WindowStyle.NoWindow);

                _grasshopperInstance = (Grasshopper.Plugin.GH_RhinoScriptInterface)Rhino.RhinoApp.GetPlugInObject("Grasshopper");

                _grasshopperInstance.DisableBanner();
                _grasshopperInstance.LoadEditor();
                _grasshopperInstance.EnableSolver();

                return true;
            }
            catch (Exception ex)
            {
                System.Windows.Forms.MessageBox.Show(ex.ToString(), "Failed to launch Rhino/GH");
                return false;
            }
        }

        
        private static void AddAssemblyResolveEvents()
        {
            try
            {
                var rhinoFolders = FindRhinoFolders(false);

                ResolveEventHandler OnRhinoCommonResolve = null;
                AppDomain.CurrentDomain.AssemblyResolve += OnRhinoCommonResolve = (sender, args) =>
                {
                    var assemblyName = new AssemblyName(args.Name).Name;

                    foreach (var rhinoFolder in rhinoFolders)
                    {
                        var dllPath = System.IO.Path.Combine(rhinoFolder, assemblyName + ".dll");
                        if (File.Exists(dllPath))
                        {
                            return Assembly.LoadFrom(dllPath);
                        }
                    }

                    return null;
                };
            }
            catch (Exception e)
            {
                System.Windows.Forms.MessageBox.Show(e.ToString(), "Exception while initializing RhinoInitializer");
            }
        }

        // adapted from https://github.com/mcneel/rhino.inside/blob/master/DotNet/RhinoInside/Resolver.cs
        private static IEnumerable<string> FindRhinoFolders(bool allowWIP)
        {
            var baseName = @"SOFTWARE\McNeel\Rhinoceros";
            using (var rhinoKey = Microsoft.Win32.Registry.LocalMachine.OpenSubKey(baseName))
            {
                if (rhinoKey == null)
                {
                    throw new Exception(@"Can't find Rhino registry info in Computer\HKEY_LOCAL_MACHINE\SOFTWARE\McNeel\Rhinoceros\");
                }
                var versionStrings = rhinoKey.GetSubKeyNames().OrderByDescending(v => v);
                foreach (var versionString in versionStrings)
                {
                    using (var installKey = rhinoKey.OpenSubKey($"{versionString}\\Install"))
                    {
                        if (installKey == null)
                        {
                            continue;
                        }
                        var buildType = installKey.GetValue("BuildType") as string;
                        if (buildType == null || (!allowWIP && buildType == "WIP"))
                        {
                            continue;
                        }

                        var foundFolders = new List<string>();

                        var coreDllPath = installKey.GetValue("CoreDllPath") as string;
                        if (coreDllPath != null && File.Exists(coreDllPath))
                        {
                            foundFolders.Add(Path.GetDirectoryName(coreDllPath));
                        }

                        var pluginsFolder = installKey.GetValue("Default Plug-ins Folder") as string;
                        if (pluginsFolder != null)
                        {
                            var grasshopperFolder = Path.Combine(pluginsFolder, "Grasshopper");
                            if (Directory.Exists(grasshopperFolder))
                            {
                                foundFolders.Add(grasshopperFolder);
                            }
                        }

                        if (foundFolders.Count >= 2)
                        {
                            return foundFolders;
                        }
                        else
                        {
                            continue;
                        }
                    }
                }
            }
            return Enumerable.Empty<string>();
        }


    }
}
